/*

intrinsics should be found in :

/usr/lib/gcc/i486-linux-gnu/4.1.2/include/xmmintrin.h

*/

#include <xmmintrin.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "test_byte_func.h"


char *prog_name;
int   check_mode=0;
int   title_mode = 0;

#define MAX_SEQ     867
#define SEQ_LENGTH  960
#ifdef COUNT
#define MAX_LOOPS  1
#else
#define MAX_LOOPS  10000
#endif

int iterations = MAX_LOOPS;
int length = SEQ_LENGTH;
int nbr_sequences = MAX_SEQ;
int method=1;
int seed=1;

unsigned long int count=0;

int (*func)(char *vA, char *vB, char *vC, int len);


char **seq;

int use_alignment = 0;

/**
 * allocate enough memory to store the initial sequences
 * and initialize the sequences with random data
 *
 */
void allocate_and_set_memory() {
  int i, j;
  size_t size = (nbr_sequences<<1)*sizeof(char *);

  if (use_alignment) {
    seq = (char **) _mm_malloc(size,16);
  } else {
    seq = (char **) malloc(size);
  }

  //printf("seq = %x\n", seq);

  for (i = 0; i <nbr_sequences; i++) {
    if (use_alignment) {
      seq[i] = (char *) _mm_malloc(length*sizeof(char),16);
    } else {
      seq[i] = calloc(length, sizeof(char));
    }
    for (j = 0; j <length; j++) {
      seq[i][j] = 1 << (rand() % 6);
      //seq[i][j] =  ((i % 2) << 1) | (j%2);
    }
  }
  while (i < ((nbr_sequences<<1)-1)) {
    if (use_alignment) {
      seq[i] = (char *) _mm_malloc(length*sizeof(char),16);
    } else {
      seq[i] = calloc(length, sizeof(char));
    }
    ++i;
  }
}

void free_memory() {
  int i;

  if (use_alignment) {
    for (i=0; i<(nbr_sequences<<1)-1; ++i) {
      _mm_free(seq[i]);
    }
    _mm_free(seq);
  } else {
    for (i=0; i<(nbr_sequences<<1)-1; ++i) {
      free(seq[i]);
    }
    free(seq);
  }
}

void align_16x_data() {
  printf("Error : sequence length must be divisible by 16");
  exit(1);
}
/**
 * usage
 *
 */
void usage() {
  printf("usage: %s --iter=n --length=n --nbr_seq=n --method=n --check\n", prog_name);
  printf("\twhere iter is the number of iterations (1..20000)\n");
  printf("\t      length is the length of a sequence (100..1000)\n");
  printf("\t      nbr_seq is the number of sequences (100..1000)\n");
  printf("\t      method used to compute parsimony (1..7)\n");
  printf("\n");
  exit(1);
}

char *options[] = {
  "--iter=",  // number of iterations
  "--length=", // sequence length
  "--nbr_seq=", // number of sequences
  "--method=", // method used to compute parsimony
  "--seed=", // seed used to initiate random number generator
  "--check", // check method
  "--title", // title of method
  NULL
};

int find_option(char *s) {
  int i=0;
  while (options[i] != NULL) {
    if (strncmp(s, options[i], strlen(options[i]))==0) return i;
    ++i;
  }
  return -1;
}

char *function_desc[] = {
  "not_available",
  "basic_function",
  "unroll_4",
  "treat_by_4",
  "vector_sse2_1_16x",
  "vector_sse2_2_16x",
  "vector_sse2_3_16x",
  "vector_sse2_4",
  "vector_sse2_5_popcount"
};

/**
 * main
 *
 */
int main(int argc, char *argv[]) {
  int i;
  int f_global=0, f_local;
  int arg, opt;
  char *s;

  prog_name = argv[0];
  arg = 1;

  while (arg < argc) {
    if (strlen(argv[arg])<3) usage();
    opt = find_option(argv[arg]);
    if (opt == -1) usage();
    s = &argv[arg][strlen(options[opt])];

    switch(opt) {
    case 0:
      iterations = atoi(s);
      if ((iterations < 1) || (iterations > 20000)) usage();
      break;

    case 1:
      length = atoi(s);
      if ((length < 100) || (length > 1000)) usage();
      break;

    case 2:
      nbr_sequences = atoi(s);
      if ((nbr_sequences < 100) || (nbr_sequences > 1000)) usage();
      break;

    case 3:
      method = atoi(s);
      if ((method < 1) || (method > 8)) usage();
      break;

    case 4:
      seed = atoi(s);
      break;

    case 5:
      check_mode = 1;
      break;

    case 6:
      title_mode = 1;
      break;

    default:
      usage();
    }
    ++arg;
  }

  if (title_mode) {
    if ((method<1) || (method>8)) usage();
    printf("%s\n",function_desc[method]);
    exit(1);
  }

  if (check_mode) {
    length = SEQ_LENGTH;
    nbr_sequences = MAX_SEQ;
  }

  // report data
  printf("iterations = %d\n", iterations);
  printf("#sequences = %d\n", nbr_sequences);
  printf("length of sequence = %d\n", length);
  printf("seed       = %d\n", seed);

  srand(seed);

  switch(method) {
  case 1:
    printf("%s\n",function_desc[1]);
    func = func_byte_basic;
    break;

  case 2:
    printf("%s\n",function_desc[2]);
    func = func_byte_unroll4;
    break;

  case 3:
    printf("%s\n",function_desc[3]);
    func = func_byte_by4;
    break;

  case 4:
    printf("%s\n",function_desc[4]);
    func = func_byte_vect1_16x;
    if ((length % 16)!=0) align_16x_data();
    use_alignment = 1;
    break;

  case 5:
    printf("%s\n",function_desc[5]);
    func = func_byte_vect2_16x;
    if ((length % 16)!=0) align_16x_data();
    use_alignment = 1;
    break;

  case 6:
    printf("%s\n",function_desc[6]);
    func = func_byte_vect3_16x;
    if ((length % 16)!=0) align_16x_data();
    use_alignment = 1;
    break;

  case 7:
    printf("%s\n",function_desc[7]);
    func = func_byte_vect4;
    use_alignment = 1;
    break;

  case 8:
    printf("%s\n",function_desc[8]);
    func = func_byte_vect5;
    use_alignment = 1;
    break;

  default:
    usage();
  }

  allocate_and_set_memory();


  if (check_mode) {

      int j = 0;
      int k = nbr_sequences;

      srand(1);

      printf("\nuse srand(1)");

      f_global=0;
      while (j < k-1) {
	f_local = (*func)(seq[j], seq[j+1], seq[k], length);
	f_global += f_local;           
	j+=2;
	++k;
      }
      printf("\nfitch's score %d", f_global);
      if (f_global == 534985)
	printf("\nOK\n");
      else
	printf("\nERROR : expected value 534985\n");

  } else { // check mode

    for (i=0; i<iterations; ++i) {
      int j = 0;
      int k = nbr_sequences;

      f_global=0;
      while (j < k-1) {
	f_local = (*func)(seq[j], seq[j+1], seq[k], length);
	f_global += f_local;           
	j+=2;
	++k;
      }
    }

    printf("\nfitch's score = %d", f_global);

  } // endif check_mode

  printf("\n");

  free_memory();

  return 0;
}


