/*
Copyright (C) 2007 Jean-Michel Richer
intrinsics should be found in :

/usr/lib/gcc/i486-linux-gnu/4.1.2/include/xmmintrin.h

compile with :

-sse2

for assembler part :

> nasm test1_asm.nasm -f elf
> gcc -c test1_func.c -msse2 -O2
> gcc -o test1.exe test1.c test1_func.o test1_asm.o

*/

#include <xmmintrin.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "test_int_func.h"


char *prog_name;


#define MAX_SEQ     867
#define SEQ_LENGTH  960
#ifdef COUNT
#define MAX_LOOPS  1
#else
#define MAX_LOOPS  10000
#endif

int iterations = MAX_LOOPS;
int length = SEQ_LENGTH;
int nbr_sequences = MAX_SEQ;
int method=1;
int seed=1;

unsigned long int count=0;

int (*func)(int *vA, int *vB, int *vC, int len);


int **seq;

int use_alignment = 0;

/**
 * allocate enough memory to store the initial sequences
 * and initialize the sequences with random data
 *
 */
void allocate_and_set_memory() {
  int i, j;
  size_t size = (nbr_sequences<<1)*sizeof(int *);

  if (use_alignment) {
    seq = (int **) _mm_malloc(size,16);
  } else {
    seq = (int **) malloc(size);
  }

  //printf("seq = %x\n", seq);

  for (i = 0; i <nbr_sequences; i++) {
    if (use_alignment) {
      seq[i] = (int *) _mm_malloc(length*sizeof(int),16);
    } else {
      seq[i] = calloc(length, sizeof(int));
    }
    for (j = 0; j <length; j++) {
      seq[i][j] = 1 << (rand() % 6);
      //seq[i][j] =  ((i % 2) << 1) | (j%2);
    }
  }
  while (i < ((nbr_sequences<<1)-1)) {
    if (use_alignment) {
      seq[i] = (int *) _mm_malloc(length*sizeof(int),16);
    } else {
      seq[i] = calloc(length, sizeof(int));
    }
    ++i;
  }
}

void free_memory() {
  int i;

  if (use_alignment) {
    for (i=0; i<(nbr_sequences<<1)-1; ++i) {
      _mm_free(seq[i]);
    }
    _mm_free(seq);
  } else {
    for (i=0; i<(nbr_sequences<<1)-1; ++i) {
      free(seq[i]);
    }
    free(seq);
  }
}

void align_16x_data() {
  printf("Error : sequence length must be divisible by 16");
  exit(1);
}
/**
 * usage
 *
 */
void usage() {
  printf("usage: %s --iter=n --length=n --nbr_seq=n --method=n\n", prog_name);
  printf("\twhere iter is the number of iterations (1..20000)\n");
  printf("\t      length is the length of a sequence (100..1000)\n");
  printf("\t      nbr_seq is the number of sequences (100..1000)\n");
  printf("\t      method used to compute parsimony (1..7)\n");
  printf("\n");
  exit(1);
}

char *options[] = {
  "--iter=",  // number of iterations
  "--length=", // sequence length
  "--nbr_seq=", // number of sequences
  "--method=", // method used to compute parsimony
  "--seed=", // seed used to initiate random number generator
  NULL
};

int find_option(char *s) {
  int i=0;
  while (options[i] != NULL) {
    if (strncmp(s, options[i], strlen(options[i]))==0) return i;
    ++i;
  }
  return -1;
}

/**
 * main
 *
 */
int main(int argc, char *argv[]) {
  int i;
  int f_global=0, f_local;
  int arg, opt;
  char *s;

  prog_name = argv[0];
  arg = 1;

  while (arg < argc) {
    if (strlen(argv[arg])<3) usage();
    opt = find_option(argv[arg]);
    if (opt == -1) usage();
    s = &argv[arg][strlen(options[opt])];
    switch(opt) {
    case 0:
      iterations = atoi(s);
      if ((iterations < 1) || (iterations > 20000)) usage();
      break;
    case 1:
      length = atoi(s);
      if ((length < 100) || (length > 1000)) usage();
      break;
    case 2:
      nbr_sequences = atoi(s);
      if ((nbr_sequences < 100) || (nbr_sequences > 1000)) usage();
      break;
    case 3:
      method = atoi(s);
      if ((method < 1) || (method > 3)) usage();
      break;
    case 4:
      seed = atoi(s);
      break;
    default:
      usage();
    }
    ++arg;
  }

  // report data
  printf("iterations = %d\n", iterations);
  printf("#sequences = %d\n", nbr_sequences);
  printf("length of sequence = %d\n", length);
  printf("seed       = %d\n", seed);

  srand(seed);

  switch(method) {
  case 1:
    printf("method     = basic function\n");
    func = func_int_basic;
    break;

  case 2:
    printf("method     = basic + unroll4\n");
    func = func_int_unroll4;
    break;

  case 3:
    printf("method     = vector SSE2 1_16x\n");
    func = func_int_vect1_16x;
    if ((length % 16)!=0) align_16x_data();
    use_alignment = 1;
    break;

  default:
    usage();
  }

  allocate_and_set_memory();

  for (i=0; i<iterations; ++i) {
    int j = 0;
    int k = nbr_sequences;

    f_global=0;
    while (j < k-1) {
      f_local = (*func)(seq[j], seq[j+1], seq[k], length);

      /*----
      {int l,f=0;
      for (l=0;l<length; ++l) {
	if ((seq[j][l] & seq[j+1][l])==0) ++f;
	printf("%3d : %3d <- %3d %3d, f= %5d\n", l, seq[j][l], 
		 seq[j+1][l], seq[k][l], f);
      }
      if (f_local != f) {
	printf("ERROR f_local = %d, f = %d\n", f_local, f);
	exit(1);
      }
      }
      ---- */

      f_global += f_local;

      
#ifdef COUNT
      { int l;
	int f=0;
	for (l=0; l<length; ++l) {
	  count += seq[k][l];
	  if ((seq[j][l] & seq[j+1][l])==0) ++f; 
	  printf("%3d: %3d <- %3d %3d, f= %5d\n", l, seq[j][l], 
		 seq[j+1][l], seq[k][l], f);
	}
      }
      break;
#endif
      
      j+=2;
      ++k;
    }
#ifdef COUNT
    printf("count = %u\n", count);
    break;
#endif
  }

  printf("fitch = %d\n", f_global);
  
  /*
#ifdef COUNT
  for (i=0; i<SEQ_LENGTH; ++i) {
    printf("%3d : %3d %3d %3d\n", i, seq[MAX_SEQ][i], seq[0][i], seq[1][i]);
  }
#endif
  */

  free_memory();

  return 0;
}

